#include "dicthiragana.h"
#include <qapplication.h>
#include <qfile.h>
#include <qmessagebox.h>
#include <qtextstream.h>
#include <qregexp.h>


DictHiragana::DictHiragana() : 
  _dicthira(QDict<QString>(17, FALSE)),
  _dictkata(QDict<QString>(17, FALSE)),
  _dictzenkaku(QDict<QString>()),
  _convmode(HIRAGANA)
{
  _dicthira.setAutoDelete(TRUE);
  _dictkata.setAutoDelete(TRUE);
}


DictHiragana::~DictHiragana() { }


void
DictHiragana::initDict(const QString& filename, QDict<QString>& dict)
{
  if ( !dict.isEmpty() ) return;
  
  QFile file(filename);
  
  if ( !file.open(IO_ReadOnly) ) {
    QMessageBox::critical(0, "File open error", "Cannot open file: " + filename,
    			  QMessageBox::Ok | QMessageBox::Default, 0);
    qApp->quit();
    return;
  }
    
  QTextStream  stream(&file);
  stream.setEncoding(QTextStream::Latin1);

  while ( !stream.eof() ) {

    QString str;
    str = QString::fromLocal8Bit(stream.readLine());
    
    if ( !str.isEmpty() && !str.contains(QRegExp("^[ \t]")) && 
	 !str.contains(QRegExp("^##")) ) {
      QString keystr, itemstr;
      int idx = str.find(QRegExp("[ \t]"));
      if (idx >= 0) {
	keystr = str.left(idx);
	
	str.remove(0, keystr.length());
	itemstr = str.replace(QRegExp("[ \t]"), "");
	
	dict.insert(keystr, new QString(itemstr));
      }
    }
  }
  
  file.close();
  return;
}


QString 
DictHiragana::convertToHira(const QString& src)
{
  if (_dicthira.isEmpty()) {
    QString dictname("dic/hiragana.dic");
    //initDict(dictname);
    initDict(dictname, _dicthira);
  }
  
  QString str_roma(src);    // ΰ
  QString dest;
  
  while ( !str_roma.isEmpty() ) {   
    uint i;
    i = (4 > str_roma.length()) ? str_roma.length() : 4;
    
    QString*  pstr;
    while (1) {
      // Ҥ餬ʸ
      pstr = _dicthira.find(str_roma.left(i));
      if (pstr) {
	dest += *pstr;
	break;
      }

      if (i == 1) {
	dest += str_roma.left(i);
	break;
      }

      i--;
    }
    
    QRegExp rxnn("nn", FALSE);
    if (i == 2 && dest.right(1) == QString::fromLocal8Bit("") 
	&& !rxnn.exactMatch(str_roma.left(i))) {
      // 'nm'ʤɤ ''ѴȤν
      str_roma.remove(0, 1);     // Σʸ
    } else if (i == 2 && str_roma[0] == str_roma[1] 
	       && !rxnn.exactMatch(str_roma.left(i))) { 
      // 'tt'ʤɤ ''Ѵʸν     
      str_roma.remove(0, 1);     // Σʸ
    } else {
      str_roma.remove(0, i);    // Ѵʬ
    }
  }
  
  return dest;
}


// convertToHira ƤӽФ塤Ǹʸ'n'ä
// '' Ѵ
QString 
DictHiragana::forceConvtToHira(const QString& src)
{
  QString dest = convertToHira(src);
  QRegExp rxn("n", FALSE);
  if ( rxn.exactMatch(dest.right(1))) {
    dest = dest.replace(dest.length() - 1, 1, QString::fromLocal8Bit("")); 
  }
  
  //qDebug("%s", dest.local8Bit().data());
  return dest;
}


QString 
DictHiragana::convertToKata(const QString& src)
{
  if (_dictkata.isEmpty()) {
    QString dictname("dic/katakana.dic");
    initDict(dictname, _dictkata);
  }

  QString dest;
  QString srchira = convertToHira(src);

  for (int i = 0; i < srchira.length(); i++) {
    QString* pstr;
    pstr = _dictkata.find( srchira.mid(i, 1) );
    if ( pstr ) {
      dest += *pstr;
    } else {
      dest += srchira.mid(i, 1);
    }
  }

  return  dest;
}


QString
DictHiragana::convertToZenkakuEisu(const QString& src)
{
  
  if (_dictzenkaku.isEmpty()) {
    QString dictname("dic/zenkakueisu.dic");
    initDict(dictname, _dictzenkaku);
  }

  QString dest;
  for (int i = 0; i < src.length(); i++) {
    QString* pstr;
    pstr = _dictzenkaku.find( src.mid(i, 1) );
    if ( pstr ) {
      dest += *pstr;
    } else {
      dest += src.mid(i, 1);
    }
  }
  
  return dest;
}


QString 
DictHiragana::forceConvtToKata(const QString& src)
{
  return convertToKata(forceConvtToHira(src));
}


QString 
DictHiragana::convertToYomi(const QString& src)
{
  switch (_convmode) {
  case HIRAGANA:
    return convertToHira(src);
    break;
    
  case KATAKANA:
    return convertToKata(src);
    break;
    
  case ZENKAKU_EISU:
    return convertToZenkakuEisu(src);
    break;
    
  default:
    Q_ASSERT(0);
    break;
  }

  return QString();
}


QString 
DictHiragana::forceConvtToYomi(const QString& src)
{
  switch (_convmode) {
  case HIRAGANA:
    return forceConvtToHira(src);
    break;
    
  case KATAKANA:
    return forceConvtToKata(src);
    break;
    
  case ZENKAKU_EISU:
    return convertToZenkakuEisu(src);
    break;
    
  default:
    Q_ASSERT(0);
    break;
  }

  return QString();
}


